//
// Created by philipp on 05.02.18.
//

#ifndef PC2CAR_TIMEDVALUE_H
#define PC2CAR_TIMEDVALUE_H

#include <NetworkingLib/Time.h>

namespace pc2car
{

template<typename T>
class TimedValue
{
public:
    using Timestamp = networking::time::TimePoint;

    TimedValue()
        : timestamp(Timestamp::min())
    {}

    TimedValue(const T & val)
    { set(val); }

    TimedValue(const T & val, const Timestamp & timestamp)
        : val(val), timestamp(timestamp)
    {}

    T get() const
    { return val; }

    void set(const T & val)
    {
        this->val = val;
        timestamp = networking::time::now();
    }

    Timestamp getTimestamp() const
    { return timestamp; }

    TimedValue & operator=(const T & val)
    { set(val); }

    explicit operator T() const
    { return get(); }

private:
    T val;
    Timestamp timestamp;
};

template<typename T>
class TimedAtomicValue
{
public:
    using Timestamp = networking::time::TimePoint;

    TimedAtomicValue()
        : timestamp(Timestamp::min())
    {}

    TimedAtomicValue(const T & val)
    { set(val); }

    T get()
    {
        std::lock_guard<std::mutex> lock{mutex};
        return val;
    }

    void set(const T & val)
    {
        std::lock_guard<std::mutex> lock{mutex};
        this->val = val;
        timestamp = networking::time::now();
    }

    Timestamp getTimestamp()
    {
        std::lock_guard<std::mutex> lock{mutex};
        return timestamp;
    }

    TimedAtomicValue & operator=(const T & val)
    { set(val); }

    explicit operator T()
    { return get(); }

    TimedValue<T> getNonAtomicCopy()
    {
        std::lock_guard<std::mutex> lock{mutex};
        return TimedValue<T>{val, timestamp};
    }

private:
    T val;
    Timestamp timestamp;
    std::mutex mutex;
};

}

#endif //PC2CAR_TIMEDVALUE_H
